#include <tchar.h>
#include <windows.h>

#include <vector>
#include <map>
#include <algorithm> // for min/max, find and sort.

#include "support.h"
#include "zipfs/zipfs.h"

#include "Scene.h"
#include "Painter.h"
#include "FontGlyph.h"

#include "Globals.h"

#include "FontEditorScene.h"

static FontEditorSceneBuilder g_fontEditorSceneBuilder;

FontEditorScene::FontEditorScene(int level) : Scene(level), m_curpos(0, 0), m_feoffset(0, 0), m_sampleoffset(10, 480 - 10 - 60 - 60 - 30) {
	m_fechar = L'あ';
	g_font.newGlyph(m_fechar); // or already exist.
	m_fecuridx = -1;
	m_femag = 1.0f;
	m_codepointi = 0;
	m_copycodei = 0;
	m_bgcodei = 0;
	m_bgindexi = -1;
	for(unsigned int i = 0; i < _countof(m_bgchars); i++) {
		m_bgchars[i] = 0;
		m_bgoffsets[i] = Point(0, 0);
	}
	m_showHandle = true;
	m_samplebuf[0] = L'\0';
	m_samplei = -1;
	m_showInfo = true;
	m_selecting = false;
	m_vibrateparam = 0;
	m_splitcodei = 0;
}

void FontEditorScene::draw(Painter &p) {
	wchar_t buf[256];
	Glyph *g = g_font.glyph(m_fechar);

	if(m_showInfo) {
		PenBrushRAII pbraii(p, 2, RGB(128, 128, 128), PS_DASH, true, HS_DIAGCROSS);

		// show current codepoint
		swprintf(buf, L"U+%04X%s", m_fechar, m_codepointi > 0 ? "?" : "");
		g_font.draw(p, Point(10, 480 - 60 - 10), 1.0f, GLYPHRAND, buf);

		// show copy-from
		if(m_copycodei > 0) {
			swprintf(buf, L"← U+%04X", m_copychar);
			g_font.draw(p, Point(10 + 40*7, 480 - 10 - 30), 0.5f, 1, buf);

			Glyph *g = g_font.glyph(m_copychar);
			if(g != NULL) {
				Point off = Point(10 + 40*7 + 20*8, 480 - 10 - 60 - 60/*buffer*/);
				off -= g->findRealTopLeft();

				{
					PenBrushRAII pbraii(p, 2, RGB(128, 128, 128), PS_DASH, true, HS_DIAGCROSS);
					p.frameRect(off.rect(Size(40, 60)));
					p.frameRect(off.rect(Size(60, 60)));
				}

				g->draw(p, off, 1.0f, 0, true);
			}
		}

		// show sample text
		if(m_samplebuf[0] != L'\0' || m_samplei != -1) {
			g_font.draw(p, m_sampleoffset, 0.5f, 1, m_samplebuf);
			g_font.draw(p, m_sampleoffset + Point(0, 30), 1.0f, 3, m_samplebuf);

			float sec = getsec();
			if(m_samplei != -1 && ((int)(sec * (1000 / 300.0f)) % 2 == 0)) {
				Size b = g_font.bounding(0.5, m_samplebuf, m_samplei);
				Point off = m_sampleoffset + b.onlycx().asPoint();
				p.line(off, off + Point(0, 30));
			}
		}

		// show split-to
		if(m_splitcodei > 0) {
			swprintf(buf, L"→ U+%04X", m_splitchar);
			g_font.draw(p, Point(10 + 40*7, 480 - 10 - 30), 0.5f, 1, buf);

			Glyph *g = g_font.glyph(m_splitchar);
			if(g != NULL) {
				Point off = Point(10 + 40*7 + 20*8, 480 - 10 - 60 - 60/*buffer*/);
				off -= g->findRealTopLeft();

				{
					PenBrushRAII pbraii(p, 2, RGB(128, 128, 128), PS_DASH, true, HS_DIAGCROSS);
					p.frameRect(off.rect(Size(40, 60)));
					p.frameRect(off.rect(Size(60, 60)));
				}

				g->draw(p, off, 1.0f, 0, true);
			}
		}

		// show debug infos
		Point dp = Point(520, 10);
		swprintf(buf, L"vk=%02X", m_vk);
		g_font.draw(p, dp, 0.3f, 0, buf);
		dp += Point(0, 20);
		swprintf(buf, L"ch=%04X", m_lastch);
		g_font.draw(p, dp, 0.3f, 0, buf);
		dp += Point(0, 20);
		if(g) {
			swprintf(buf, L"ps=%d", g->pointlen());
			g_font.draw(p, dp, 0.3f, 0, buf);
			dp += Point(0, 20);

			swprintf(buf, L"o(%d,%d)", m_feoffset.x(), m_feoffset.y());
			g_font.draw(p, dp, 0.3f, 0, buf);
			dp += Point(0, 20);

			swprintf(buf, L"%.1f%%", m_femag * 100);
			g_font.draw(p, dp, 0.3f, 0, buf);
			dp += Point(0, 30);

			for(unsigned int i = 0; i < _countof(m_bgchars); i++) {
				if(m_bgchars[i] != 0) {
					swprintf(buf, L"b%d(%d,%d)", i, m_bgoffsets[i].x(), m_bgoffsets[i].y());
					g_font.draw(p, dp, 0.3f, 0, buf);
					dp += Point(0, 20);
				}
			}

			Point curpos = (m_curpos - m_feoffset) / m_femag;
			swprintf(buf, L"m(%d,%d)", curpos.x(), curpos.y());
			g_font.draw(p, dp, 0.3f, 0, buf);
			dp += Point(0, 20);
		}

		if(m_samplei != -1 || 0 < wcslen(m_samplebuf)) {
			swprintf(buf, L"s(%d,%d)", m_sampleoffset.x(), m_sampleoffset.y());
			g_font.draw(p, dp, 0.3f, 0, buf);
			dp += Point(0, 20);
		}
	}

	// show glyph layers
	if(g) {
		// show bg glyphs
		for(unsigned int i = 0; i < _countof(m_bgchars); i++) {
			if(m_bgchars[i] != 0) {
				Glyph *bg = g_font.glyph(m_bgchars[i]);
				if(bg) {
					PenBrushRAII pbraii(p, 2, (m_bgindexi == static_cast<int>(i)) ? RGB(128, 128, 128) : RGB(64, 64, 64), PS_SOLID, false, 0);
					bg->draw(p, m_feoffset + m_bgoffsets[i] * m_femag, m_femag, m_vibrateparam, true);
				}
			}

			if(m_showInfo && (m_bgchars[i] != 0 || m_bgcodei == 5 || (m_bgindexi == static_cast<int>(i) && m_bgcodei > 0))) {
				PenBrushRAII pbraii(p, 2, RGB(128, 128, 128), PS_DASH, true, HS_DIAGCROSS);
				swprintf(buf, L"%d@U+%04X%s", i, m_bgchars[i], (m_bgcodei == 5 || m_bgindexi == static_cast<int>(i)) ? L"?" : L"");
				g_font.draw(p, Point(10 + 40*7 + (i / 2 * 13*8), 480 - 10 - 30 - 40 + (i % 2 * 20)), 0.3f, 1, buf);
			}
		}

		// show glyph guidelines
		// frames are not depend with Glyph, but for understandable.
		{
			PenBrushRAII pbraii(p, 2, RGB(128, 128, 128), PS_DASH, true, HS_DIAGCROSS);
			p.frameRect(m_feoffset.rect(Size(40, 60)));
			p.frameRect(m_feoffset.rect(Size(60, 60)));
			p.frameRect(m_feoffset.rect(Size(40, 60) * m_femag));
			p.frameRect(m_feoffset.rect(Size(60, 60) * m_femag));
		}

		// show current glyph
		g->draw(p, m_feoffset, m_femag, m_vibrateparam, true, m_showHandle, m_fecuridx, &m_curpos);
	}

	if(m_selecting) {
		PenBrushRAII pbraii(p, 1, RGB(255, 0, 0), PS_SOLID, false, 0);
		p.frameRect(Rect(m_selstart * m_femag + m_feoffset, (m_curpos - m_feoffset) / m_femag * m_femag + m_feoffset));
	}
	if(m_showHandle && m_selpoints.size() > 0) {
		PenBrushRAII pbraii(p, 1, RGB(0, 255, 0), PS_SOLID, false, 0);
		unsigned int pl = g->pointlen();
		for(std::vector<unsigned int>::iterator it = m_selpoints.begin(); it != m_selpoints.end(); it++) {
			if(*it < pl) {
				const GlyphPoint *gp = g->point(*it);
				if(gp) {
					p.frameRect((gp->first * m_femag + m_feoffset).box(3));
				}
			}
		}
	}
}

void FontEditorScene::button(enum MouseButton btn, Point curpos, bool isdown) {
	m_curpos = curpos;

	if(!isdown) {
		if(btn == LBUTTON && m_selecting) {
			m_selecting = false;
		}

		return;
	}

	if(m_button == LBUTTON) {
		if(isKey(VK_SHIFT)) {
			m_selstart = (curpos - m_feoffset) / m_femag;
			m_selecting = true;
			drag(btn, curpos); // delegate
		} else {
			Glyph *g = g_font.glyph(m_fechar);
			if(g) {
				Point curpos = (m_curpos - m_feoffset) / m_femag;
				if(m_fecuridx == -1) {
					// new point
					g->insPoint(g->pointlen(), curpos, true);
					if(isKey(VK_CONTROL)) {
						m_fecuridx = -1;
					} else {
						m_fecuridx = g->pointlen();
					}
				} else if(m_fecuridx == static_cast<int>(g->pointlen())) {
					// new point (continue)
					g->insPoint(g->pointlen(), curpos, false);
					if(isKey(VK_CONTROL)) {
						m_fecuridx = -1;
					} else {
						m_fecuridx = g->pointlen();
					}
				} else {
					// extend(next is newpoint) or set(new is not)
					const GlyphPoint *gp = g->point(m_fecuridx);
					g->setPoint(m_fecuridx, curpos, gp->second);
					if(m_fecuridx + 1 == static_cast<int>(g->pointlen())) {
						// extend last segment
						if(isKey(VK_CONTROL)) {
							m_fecuridx = -1;
						} else {
							m_fecuridx = m_fecuridx + 1;
						}
					} else if (g->point(m_fecuridx + 1)->second == true) { // next is newpoint
						// extend current segment
						if(isKey(VK_CONTROL)) {
							m_fecuridx = -1;
						} else {
							g->insPoint(m_fecuridx + 1, curpos, false); // XXX: insert at place time, not at this moment...
							m_fecuridx = m_fecuridx + 1;
						}
					} else {
						// set
						m_fecuridx = -1;
					}
				}
			}
		}
	}
	if(m_button == MBUTTON) {
		if(isKey(VK_SHIFT) && isKey(VK_CONTROL)) {
			if(m_samplei != -1) {
				m_samplemoveoff = m_curpos - m_sampleoffset;
			} else {
				for(unsigned int i = 0; i < _countof(m_bgchars); i++) {
					m_bgofforgs[i] = m_curpos - m_bgoffsets[i];
				}
			}
		} else if(isKey(VK_SHIFT) && !isKey(VK_CONTROL)) {
			m_offorg = m_curpos - m_feoffset;
		} else if(!isKey(VK_SHIFT) && isKey(VK_CONTROL)) {
			if(m_selpoints.size() == 0) {
				Point curpos = (m_curpos - m_feoffset) / m_femag;
				Glyph *g = g_font.glyph(m_fechar);
				m_realtlorg = curpos - g->findRealTopLeft();
			} else {
				Point curpos = (m_curpos - m_feoffset) / m_femag;
				m_selmoveorg = curpos;
			}
		}
		// delegate
		drag(btn, curpos);
	}
	if(m_button == RBUTTON) {
		if(!isKey(VK_SHIFT)) {
			m_selpoints.clear();
		}

		Glyph *g = g_font.glyph(m_fechar);
		if(g) {
			Point curpos = (m_curpos - m_feoffset) / m_femag;
			for(unsigned int i = 0; i < g->pointlen(); i++) {
				if((g->point(i)->first * m_femag).box(3).has(m_curpos - m_feoffset)) { // note: "m_curpos" not "curpos".
					if(isKey(VK_SHIFT)) {
						std::vector<unsigned int>::iterator it = std::find(m_selpoints.begin(), m_selpoints.end(), i);
						if(it == m_selpoints.end()) {
							m_selpoints.push_back(i);
							std::sort(m_selpoints.begin(), m_selpoints.end()); // keep sorted!
						} else {
							m_selpoints.erase(it);
						}
					} else {
						if(m_fecuridx == static_cast<int>(i)) {
							// continue(next is newpoint) or split(next is not)
							g->insPoint(i + 1, curpos, false);
							m_fecuridx = i + 1;
							return;
						} else {
							// edit this
							m_fecuridx = static_cast<int>(i);
							return;
						}
					}
				}
			}
			m_fecuridx = -1;
		}
	}
}

void FontEditorScene::drag(enum MouseButton btn, Point curpos) {
	m_curpos = curpos;

	if(m_button == LBUTTON && m_selecting) {
		Rect rc(m_selstart, (curpos - m_feoffset) / m_femag);
		Glyph *g = g_font.glyph(m_fechar);
		m_selpoints.clear();
		unsigned int np = g->pointlen();
		for(unsigned int i = 0; i < np; i++) {
			if(rc.has(g->point(i)->first)) {
				m_selpoints.push_back(i);
			}
		}
	}
	if(m_button == MBUTTON) {
		// Mbutton pressing
		Point curpos = m_curpos - m_feoffset;
		if(isKey(VK_CONTROL) && isKey(VK_SHIFT)) {
			if(m_samplei != -1) {
				// sample text offset
				m_sampleoffset = m_curpos - m_samplemoveoff;
			} else {
				// bg offset
				for(unsigned int i = 0; i < _countof(m_bgchars); i++) {
					if(m_bgindexi != -1 && m_bgindexi != static_cast<int>(i)) {
						continue;
					}
					m_bgoffsets[i] = (m_curpos - m_bgofforgs[i]) / m_femag;
				}
			}
		} else if(isKey(VK_CONTROL) && !isKey(VK_SHIFT)) {
			if(m_selpoints.size() == 0) {
				// glyph points offset
				g_font.glyph(m_fechar)->justify(curpos / m_femag - m_realtlorg);
			} else {
				// selected glyph points offset
				Point curposg = curpos / m_femag;
				Point cdiff = curposg - m_selmoveorg;
				Glyph *g = g_font.glyph(m_fechar);
				unsigned int pl = g->pointlen();
				for(std::vector<unsigned int>::iterator it = m_selpoints.begin(); it != m_selpoints.end(); it++) {
					if(*it < pl) {
						const GlyphPoint *gp = g->point(*it);
						if(gp) {
							g->setPoint(*it, gp->first + cdiff, gp->second);
						}
					}
				}
				m_selmoveorg = curposg;
			}
		} else if(isKey(VK_SHIFT) && !isKey(VK_CONTROL)) {
			// view offset
			m_feoffset = m_curpos - m_offorg;
		} else {
			if(m_fecuridx != -1) {
				// cut/melding current point
				Glyph *g = g_font.glyph(m_fechar);
				const GlyphPoint *gp = g->point(m_fecuridx);
				g->setPoint(m_fecuridx, gp->first, !gp->second);
			} else {
				// zooming
				if((float)curpos.x()/curpos.y() > (float)40/60) {
					m_femag = (float)curpos.x() / 40;
				} else {
					m_femag = (float)curpos.y() / 60;
				}
			}
		}
	}
}

void FontEditorScene::wheel(Point curpos, float delta) {
	m_femag += delta * 0.25f;
}

void FontEditorScene::key(UINT vk, bool isdown) {
	if(!isdown) {
		return;
	}

	/*
	1: 100%
	0: offset (0,0) with all bg
	B: BG specify ("B"->"*": remove all bg)
	C: Append(a.k.a. Copy) from...
	F: Flip Vert (+Shift=Horz)
	H: Show/Hide handle
	I: Show/Hide info
	K: Apply(Make permanent) current zoom
	L: Load (+Shift=txt)
	M: Justify
	P: Split (and appending) select points to...
	S: Save (+Shift=txt)
	T: Sample text (T twice=clear current text)
	U: Codepoint...
	V: Vibrate preview (0->3->1->0->...)
	Left: Previous codepoint (+Ctrl=decr)
	Right: Next codepoint (+Ctrl=incr)
	Bksp: (Sample text: erase left char of caret)
	Del: Delete selected point (+Ctrl=DELETE ALL POINT)
	Esc: Cancel current selection, codepoint input, sample input, ...
	Return: (Copy from: accept, Sample text: complete)
	Ins: (Sample text: insert current codepoint char)
	*/

	if(m_vk == 'S' && isKey(VK_CONTROL)) {
		if(isKey(VK_SHIFT)) {
			g_font.dumpText();
		} else {
			g_font.dump();
		}
		return;
	}
	if(m_vk == 'L' && isKey(VK_CONTROL)) {
		if(isKey(VK_SHIFT)) {
			long fontsize;
			BYTE *fontdata = unzReadWhole(g_unz, "font.txt", &fontsize);
			if(fontdata != NULL) {
				g_font.loadText(fontdata);
				free(fontdata);
			}
		} else {
			long fontsize;
			BYTE *fontdata = unzReadWhole(g_unz, "font.bin", &fontsize);
			if(fontdata != NULL) {
				g_font.load(fontdata);
				free(fontdata);
			}
		}
		Glyph *g = g_font.glyph(m_fechar);
		if(!g) {
			// if current char is not defined, get first "ch".
			m_fechar = g_font.nextch(0);
		}
		return;
	}
	if(m_vk == 'M' && isKey(VK_CONTROL)) {
		g_font.glyph(m_fechar)->justify(Point(0, 0));
		return;
	}
	if(m_vk == '1' && isKey(VK_CONTROL)) {
		m_femag = 1.0f;
		return;
	}
	if(m_vk == 'U' && isKey(VK_CONTROL)) {
		m_codepointi = 4;
		m_fechar = 0;
		return;
	}
	if(m_vk == 'C' && isKey(VK_CONTROL)) {
		m_copycodei = 5;
		m_copychar = 0;
		return;
	}
	if(m_vk == 'B' && isKey(VK_CONTROL)) {
		if(m_bgcodei == 0) {
			m_bgcodei = 5;
			//m_bgchar = 0;
		} else {
			if(m_bgcodei == 5) {
				for(unsigned int i = 0; i < _countof(m_bgchars); i++) {
					if(m_bgchars[i] == 0) {
						m_bgchars[i] = m_fechar;
						m_bgcodei = 0;
						m_bgindexi = -1;
						break;
					}
				}
			} else if(m_bgindexi != -1) {
				m_bgchars[m_bgindexi] = m_fechar;
				m_bgcodei = 0;
				m_bgindexi = -1;
			}
		}
		return;
	}
	if(m_vk == '0' && isKey(VK_CONTROL)) {
		if(m_bgcodei == 0) {
			m_feoffset = Point(0, 0);
		} else if(m_bgcodei == 5) {
			for(unsigned int i = 0; i < _countof(m_bgchars); i++) {
				m_bgoffsets[i] = Point(0, 0);
			}
		} else {
			m_bgoffsets[m_bgindexi] = Point(0, 0);
		}
		return;
	}
	if(m_vk == 'H' && isKey(VK_CONTROL)) {
		m_showHandle = !m_showHandle;
		return;
	}
	if(m_vk == 'K' && isKey(VK_CONTROL)) {
		Glyph *glyph = g_font.glyph(m_fechar);
		glyph->scale(m_femag);
		return;
	}
	if(m_vk == 'T' && isKey(VK_CONTROL)) {
		if(m_samplei == -1) {
			m_samplei = wcslen(m_samplebuf);
		} else {
			m_samplebuf[0] = L'\0';
			m_samplei = 0;
		}
		return;
	}
	if(m_vk == 'I' && isKey(VK_CONTROL)) {
		m_showInfo = !m_showInfo;
		return;
	}
	if(m_vk == 'F' && isKey(VK_CONTROL)) {
		Glyph *g = g_font.glyph(m_fechar);
		Point tl = g->findRealTopLeft();
		Size realsz = g->bounding() - tl.asSize();
		unsigned int pl = g->pointlen();
		for(unsigned int i = 0; i < pl; i++) {
			const GlyphPoint *gp = g->point(i);
			Point gpp = gp->first;
			if(isKey(VK_SHIFT)) {
				// horz flip
				gpp.setx(tl.x() + (realsz.cx() - (gpp.x() - tl.x())));
			} else {
				// vert flip
				gpp.sety(tl.y() + (realsz.cy() - (gpp.y() - tl.y())));
			}
			g->setPoint(i, gpp, gp->second);
		}
		return;
	}
	if(m_vk == 'V' && isKey(VK_CONTROL)) {
		switch(m_vibrateparam) {
		case 0: m_vibrateparam = 3; break;
		case 3: m_vibrateparam = 1; break;
		case 1: m_vibrateparam = 0; break;
		default: m_vibrateparam = 0; break; // ?
		}
	}
	if(m_vk == 'P' && isKey(VK_CONTROL)) {
		if(m_selpoints.size() > 0) {
			m_splitcodei = 5;
			m_splitchar = 0;
		}
		return;
	}
	if(m_vk == VK_LEFT) {
		if(m_samplei != -1) {
			m_samplei = std::max(0, m_samplei - 1);
		} else {
			if(isKey(VK_CONTROL)) {
				m_fechar--;
				g_font.newGlyph(m_fechar); // or exists.
			} else {
				m_fechar = g_font.prevch(m_fechar);
			}
		}
		return;
	}
	if(m_vk == VK_RIGHT) {
		if(m_samplei != -1) {
			m_samplei = std::min(static_cast<int>(wcslen(m_samplebuf)), m_samplei + 1);
		} else {
			if(isKey(VK_CONTROL)) {
				m_fechar++;
				g_font.newGlyph(m_fechar); // or exists.
			} else {
				m_fechar = g_font.nextch(m_fechar);
			}
		}
		return;
	}
	if(m_vk == VK_BACK) {
		if(m_samplei != -1) {
			if(m_samplei > 0) {
				int len = wcslen(m_samplebuf + m_samplei);
				memmove(m_samplebuf + m_samplei - 1, m_samplebuf + m_samplei, (len + 1) * 2); // +1 = L'\0'
				m_samplei--;
			}
		}
		return;
	}
	if(m_vk == VK_DELETE) {
		Glyph *glyph = g_font.glyph(m_fechar);
		if(isKey(VK_CONTROL)) {
			// Ctrl+Del
			glyph->clearPoints();
		} else {
			if(m_selpoints.size() > 0) {
				// remove selected points, also clear points selection.
				// XXX: assuming m_selpoints is sorted ascended!
				//      removing point must be from tail. (or indexes changed and die.)
				while(m_selpoints.size() > 0) {
					glyph->delPoint(m_selpoints.back());
					m_selpoints.pop_back();
				}
			} else {
				// remove current point if selected (by RButton).
				if(m_fecuridx != -1 && static_cast<unsigned int>(m_fecuridx) < glyph->pointlen()) {
					bool stop = glyph->point(m_fecuridx)->second;
					glyph->delPoint(m_fecuridx);
					if(!stop && (m_fecuridx == static_cast<int>(glyph->pointlen()) || glyph->point(m_fecuridx)->second == true)) {
						// deleting segment last point(but not first): select previous point to easy to delete.
						m_fecuridx--;
						return;
					}
				}
			}
		}
		m_fecuridx = -1;
		return;
	}
	if(m_vk == VK_ESCAPE) {
		// cancelling Unicodepoint inputting
		m_codepointi = 0;
		m_copycodei = 0;
		if(m_bgindexi != -1 && g_font.glyph(m_bgchars[m_bgindexi]) == NULL) {
			m_bgchars[m_bgindexi] = L'\0';
		}
		m_bgcodei = 0;
		m_bgindexi = -1;
		m_splitcodei = 0;
		// cancelling sample text inputting
		m_samplei = -1;
		return;
	}
	if(m_vk == VK_RETURN) {
		if(m_copycodei == 1) {
			m_copycodei = 0;

			Glyph *sg = g_font.glyph(m_copychar); // must exist.
			Glyph *dg = g_font.glyph(m_fechar); // must exist.
			// dg->setGlyph(*sg); // do not override...
			// ...append instead. can override with "Ctrl+Del" and this "C".
			m_selpoints.clear(); // also selects point that copied here.
			unsigned int spl = sg->pointlen();
			unsigned int dpl = dg->pointlen();
			for(unsigned int i = 0; i < spl; i++) {
				const GlyphPoint *gp = sg->point(i);
				dg->insPoint(dpl + i, gp->first, gp->second);
				m_selpoints.push_back(dpl + i);
			}
		}
		if(m_samplei != -1) {
			m_samplei = -1;
		}
		if(m_splitcodei == 1) {
			m_splitcodei = 0;

			Glyph *sg = g_font.glyph(m_fechar); // must exist.
			Glyph *dg = g_font.glyph(m_splitchar); // must exist.
			// append. can override with "Ctrl+Del" on dst char and this "P" on src char.
			unsigned int dpl = dg->pointlen();
			for(std::vector<unsigned int>::iterator it = m_selpoints.begin(); it != m_selpoints.end(); it++) {
				const GlyphPoint *gp = sg->point(*it);
				if(gp) {
					dg->insPoint(dpl++, gp->first, gp->second);
				}
			}
		}
	}
	if(m_vk == VK_INSERT) {
		if(m_samplei != -1) {
			// inserting current glyph to sample string.
			int len = wcslen(m_samplebuf + m_samplei);
			memmove(m_samplebuf + m_samplei + 1, m_samplebuf + m_samplei, (len + 1) * 2); // +1 = L'\0'
			m_samplebuf[m_samplei] = m_fechar;
			m_samplei++;
		}
	}
}

void FontEditorScene::moji(TCHAR ch) {
	m_lastch = ch; // debug

	if(m_codepointi > 0 || m_copycodei > 0 || m_bgcodei > 0 || m_splitcodei > 0) {
		if(_istxdigit(ch)) {
			int i;
			if(ch <= _T('9')) {
				i = ch - _T('0');
			} else if(ch <= _T('F')) {
				i = ch - _T('A') + 10;
			} else if(ch <= _T('f')) {
				i = ch - _T('a') + 10;
			} else {
				return;
			}
			if(m_codepointi > 0) {
				m_fechar = m_fechar * 16 + i;
				m_codepointi--;
				if(m_codepointi == 0) {
					g_font.newGlyph(m_fechar); // or already exist.
				}
			}
			if(m_copycodei > 1) {
				m_copychar = m_copychar * 16 + i;
				m_copycodei--;
				if(m_copycodei == 1) {
					Glyph *g = g_font.glyph(m_copychar);
					if(!g) {
						// glyph not found; force cancel.
						m_copycodei = 0;
					}
				}
			}
			if(m_bgcodei > 0) {
				if(m_bgcodei == 5) {
					m_bgindexi = i;
					m_bgcodei--;
					if(static_cast<int>(_countof(m_bgchars)) <= m_bgindexi) {
						m_bgcodei = 0;
						m_bgindexi = -1;
					} else {
						// do not clear bgchar, it used as selecting bg when moving bg.
						// instead, clear when first char entry.
						//m_bgchars[m_bgindexi] = 0;
					}
				} else {
					if(m_bgcodei == 4) {
						m_bgchars[m_bgindexi] = 0;
					}
					m_bgchars[m_bgindexi] = m_bgchars[m_bgindexi] * 16 + i;
					m_bgcodei--;
					if(m_bgcodei == 0) {
						if(g_font.glyph(m_bgchars[m_bgindexi]) == NULL) {
							// cannot specify non-defined-char as bgchar.
							m_bgchars[m_bgindexi] = 0;
						}
						m_bgindexi = -1;
					}
				}
			}
			if(m_splitcodei > 1) {
				m_splitchar = m_splitchar * 16 + i;
				m_splitcodei--;
				if(m_splitcodei == 1) {
					g_font.newGlyph(m_splitchar); // or exists.
				}
			}
		}
		if(ch == '*') {
			for(unsigned int i = 0; i < _countof(m_bgchars); i++) {
				m_bgchars[i] = 0;
			}
			m_bgcodei = 0;
			m_bgindexi = -1;
		}
		return;
	}
	if(0x20 <= ch) {
		if(m_samplei != -1) {
			int len = wcslen(m_samplebuf + m_samplei);
			memmove(m_samplebuf + m_samplei + 1, m_samplebuf + m_samplei, (len + 1) * 2); // +1 = L'\0'
			m_samplebuf[m_samplei] = ch;
			m_samplei++;
		} else {
			//debug(_T("%d incoming"), ch);
			m_fechar = ch;
			g_font.newGlyph(m_fechar); // or exist.
		}
	}
}
