#include <tchar.h>
#include <windows.h>

#include <math.h>

#include <vector>
#include <map>

#include "support.h"
#include "zipfs/zipfs.h"
#include "wavemix/wavemix.h"
#include "zipwave.h"

#include "Scene.h"
#include "Painter.h"
#include "FontGlyph.h"

#include "Globals.h"

#include "KillBossGameScene.h"


enum KillBossGlyphs {
	KBG_BOSS1 = 0xE040,
	KBG_BOSS2,
	KBG_BOSS3,
	KBG_BOSSBULLET = 0xE043,
	KBG_ME1 = 0xE044,
	KBG_ME2,
	KBG_MYBULLET = 0xE046,
	KBG_POMF1 = 0xE047,
	KBG_POMF2,
	KBG_GAUGEFRAME = 0xE049,
	KBG_GAUGEFILL = 0xE04A,
};

const int MYY = 370;

static KillBossGameSceneBuilder g_killBossGameSceneBuilder;

KillBossGameScene::KillBossGameScene(int level) : Scene(level) {
	m_bossch = KBG_BOSS1 + level - 1;
	m_maxbosslife = level * 2 + 1;
	m_bosslife = m_maxbosslife;
	m_bosspos = Point(randr(40, 600), randr(0, 150));
	m_myx = 320 - g_font.glyph(KBG_ME1)->bounding().cx() / 2; // center
	m_pretargetsec = 0.0f;
	m_prebossbulletsec = 0.0f;
	m_bosspomf = false;
	m_mypomf = false;
	m_bossmovetime = 1.0f - 0.2f * level;

	newBossTarget();

	m_curpos = Point(320, 0);

	m_wbgm = unzLoadWave("st008.wav");
	m_wbossbullet = unzLoadWave("se_maoudamashii_battle_gun05.wav");
	m_wmybullet = unzLoadWave("se_maoudamashii_battle_gun03.wav");
	m_wbomb = unzLoadWave("Blast-SoundBible.com-2068539061.wav");

	m_ticked = false;
}

KillBossGameScene::~KillBossGameScene() {
	wmKillWave(m_wbgm);

	wmUnloadWave(m_wbgm);
	wmUnloadWave(m_wbossbullet);
	wmUnloadWave(m_wmybullet);
	wmUnloadWave(m_wbomb);
}

static void forwardBullets(std::vector<std::pair<Point, Point> > &vec, wchar_t ch) {
	Size bb = g_font.glyph(ch)->bounding();
	Rect disp(-bb.asPoint(), Point(640, 480));
retry:
	for(std::vector<std::pair<Point, Point> >::iterator it = vec.begin(); it != vec.end(); it++) {
		it->first += it->second;
		if(!disp.has(it->first)) {
			vec.erase(it);
			goto retry;
		}
	}
}

void KillBossGameScene::tick() {
	if(!m_ticked) {
		m_ticked = true;
		wmPlayWave(m_wbgm);
	}

	if(!m_bosspomf) {
		if(m_bossmovetime < getsec() - m_pretargetsec) {
			m_pretargetsec = getsec();
			newBossTarget();
		}

		if(0.5f < getsec() - m_prebossbulletsec) { // TODO: level
			m_prebossbulletsec = getsec();

			Glyph *g1 = g_font.glyph(m_bossch);
			Glyph *g2 = g_font.glyph(KBG_BOSSBULLET);
			Point pos = m_bosspos + Point((g1->bounding().cx() - g2->bounding().cx()) / 2, 50);
			m_bossbullets.push_back(std::pair<Point,Point>(pos, Point(0, 10)));
			wmPlayWave(m_wbossbullet);
		}

		m_bosspos = Point((m_bosspos * 5 + m_bosstarget) / 6);

		if(!m_mypomf) {
			Size gbbsz = g_font.glyph(KBG_BOSSBULLET)->bounding();
			Rect myrc = (Point(m_myx, MYY) + gbbsz.asPoint().onlyx() / 2).rect(g_font.glyph(KBG_ME1)->bounding() - gbbsz.onlycx()); // note: inverted signedness to make avoid easy
			for(std::vector<std::pair<Point, Point> >::iterator it = m_bossbullets.begin(); it != m_bossbullets.end(); it++) {
				if(myrc.has(it->first)) {
					m_mypomf = true;
					wmPlayWave(m_wbomb);
					m_bossbullets.erase(it);
					// note: dont make SR_FAILED here, delay until no player bullets exists.
					break;
				}
			}
		}

		// note: player invulnerable when boss is pomf-ing, but boss is not when player is pomf-ing.
		Glyph *g = g_font.glyph(m_bossch);
		Point rtl = g->findRealTopLeft();
		Size sz = g->bounding() - rtl.asSize();
		Size gmbsz = g_font.glyph(KBG_MYBULLET)->bounding();
		Rect bossrc = (m_bosspos + rtl - gmbsz.asPoint().onlyx() / 2).rect(sz + gmbsz.onlycx());
		for(std::vector<std::pair<Point, Point> >::iterator it = m_mybullets.begin(); it != m_mybullets.end(); it++) {
			if(bossrc.has(it->first)) {
				m_bosslife--;
				if(m_bosslife <= 0) {
					m_bosspomf = true;
					wmPlayWave(m_wbomb);
					setResult(SR_SUCCESS); // immediately success.
				}
				m_mybullets.erase(it);
				break;
			}
		}

		if(m_mybullets.size() == 0 && m_mypomf) {
			// make fail only when boss is alive and player is dead
			if(result() == SR_PLAYING)
			{
				setResult(SR_FAILED);
			}
		}
	}

	forwardBullets(m_bossbullets, KBG_BOSSBULLET);
	forwardBullets(m_mybullets, KBG_MYBULLET);

	if(!m_mypomf) {
		int mycenter = g_font.glyph(KBG_ME1)->bounding().cx() / 2;
		m_myx = m_myx + std::min(std::max(static_cast<int>(m_curpos.x()) - mycenter - m_myx, -5), 5);
	}
}

void KillBossGameScene::draw(Painter &p) {
	Glyph *g;

	if(!m_bosspomf) {
		g = g_font.glyph(m_bossch);
	} else {
		g = g_font.glyph(KBG_POMF1 + static_cast<int>(getsec() / 0.3f) % 2);
	}
	g->draw(p, m_bosspos, 1.0f, GLYPHRAND);

	if(!m_mypomf) {
		g = g_font.glyph(KBG_ME1 + static_cast<int>(getsec() / 0.3f) % 2);
		g->draw(p, Point(m_myx, MYY), 1.0f, GLYPHRAND);
	} else {
		g = g_font.glyph(KBG_POMF1 + static_cast<int>(getsec() / 0.3f) % 2);
		g->draw(p, Point(m_myx, MYY), 0.5f, GLYPHRAND);
	}

	g = g_font.glyph(KBG_BOSSBULLET);
	for(std::vector<std::pair<Point, Point> >::iterator it = m_bossbullets.begin(); it != m_bossbullets.end(); it++) {
		g->draw(p, it->first, 1.0f, GLYPHRAND);
	}

	g = g_font.glyph(KBG_MYBULLET);
	for(std::vector<std::pair<Point, Point> >::iterator it = m_mybullets.begin(); it != m_mybullets.end(); it++) {
		g->draw(p, it->first, 1.0f, GLYPHRAND);
	}

	// drawing boss life gauge.
	g_font.draw(p, Point(10, 10), 0.5f, GLYPHRAND, L"BOSS");

	g = g_font.glyph(KBG_GAUGEFRAME);
	g->draw(p, Point(100, 10), 1.0f, GLYPHRAND);

	g = g_font.glyph(KBG_GAUGEFILL);
	Point pt(100, 10);
	int max = 52 * m_bosslife / m_maxbosslife;
	for(int i = 0; i < max; i++) {
		g->draw(p, pt, 1.0f, GLYPHRAND);
		pt.setx(pt.x() + g->bounding().cx());
	}
}

void KillBossGameScene::button(enum MouseButton btn, Point curpos, bool isdown) {
	if(btn == LBUTTON && isdown && !m_mypomf) {
		if(m_mybullets.size() < 2) {
			Glyph *g1 = g_font.glyph(KBG_ME1);
			Glyph *g2 = g_font.glyph(KBG_MYBULLET);
			int x = m_myx + (g1->bounding().cx() - g2->bounding().cx()) / 2;
			m_mybullets.push_back(std::pair<Point,Point>(Point(x, MYY), Point(0, -15)));
			wmPlayWave(m_wmybullet);
		}
	}

	hover(curpos); // delegate
}

void KillBossGameScene::drag(enum MouseButton btn, Point curpos) {
	hover(curpos); // delegate
}

void KillBossGameScene::hover(Point curpos) {
	m_curpos = curpos;
}

void KillBossGameScene::newBossTarget() {
	Size sz = g_font.glyph(m_bossch)->bounding();
	m_bosstarget = Point(randr(40, 600 - sz.cx()), randr(0, 150));
}
