#include <tchar.h>
#include <windows.h>

#include <math.h>

#include <vector>
#include <map>

#include "support.h"
#include "zipfs/zipfs.h"
#include "wavemix/wavemix.h"
#include "zipwave.h"

#include "Scene.h"
#include "Painter.h"
#include "FontGlyph.h"

#include "Globals.h"

#include "MainGameScene.h"
#include "EndrollScene.h"
#include "TitleScene.h"

enum MainGameGlyphs {
	MGG_ELEVATOR0 = 0xE010,
	MGG_ELEVATOR1,
	MGG_ELEVATOR2,
	MGG_ELEVATOR3,
	MGG_ELEVATOR4,
	MGG_ELEVATOR5,
	MGG_LEVELUP = 0xE016,
	MGG_LUPSHOUT1 = 0xE017,
	MGG_LUPSHOUT2,
	MGG_OK1 = 0xE019,
	MGG_OK2,
	MGG_BAD1 = 0xE01B,
	MGG_BAD2,
	MGG_PATCH1 = 0xE01D,
	MGG_PATCH2,

	MGG_COUNTDOWN1 = 0xE030,
	MGG_COUNTDOWN2,
	MGG_COUNTDOWN3,
	MGG_COUNTDOWN4,
	MGG_COUNTDOWN5,
	MGG_COUNTDOWN6,
	MGG_COUNTDOWN7,
	MGG_COUNTDOWN8,
};

static const Rect g_elevatorRects[] = {
	Rect(287, 137, 389, 480),
	Rect(192, 137, 463, 480),
	Rect( 89, 137, 551, 480),
	Rect( 44,  74, 599, 480),
	Rect( 23,  41, 616, 480),
};

static const Point g_patchpoints[4] = {
	Point(  4, 157),
	Point( 21, 331),
	Point(540, 191),
	Point(555, 362),
};

static const int g_patchglyph[4] = {0, 1, 1, 0};

static const float ZOOM_DURATION = 0.3f;
static const float LIMIT_BUFFER = 0.3f;

static MainGameSceneBuilder g_mainGameSceneBuilder;

MainGameScene::MainGameScene(int level) : Scene(level) {
	m_level = 1;
	m_stageno = 0;
	m_failed = 0;
	m_gamescene = NULL;
	m_deadlinesec = 0.0f;
	m_limitover = isKey(VK_CONTROL);

	m_ok = unzLoadWave("muci_fan_05.wav");
	m_bad = unzLoadWave("Buzzer2.wav");
	m_pregame = unzLoadWave("muci_action_03.wav");
	m_levelup = unzLoadWave(/*"accordion.wav"*/"se_maoudamashii_jingle12.wav");
	m_endgame = unzLoadWave("Untitled.wav");
	m_tick = unzLoadWave("tick.wav");
	m_tock = unzLoadWave("tock.wav");
	m_bomb = unzLoadWave("Bomb_Exploding-Sound_Explorer-68256487.wav");

	newstate(GS_PREGAME);
}

MainGameScene::~MainGameScene() {
	wmUnloadWave(m_ok);
	wmUnloadWave(m_bad);
	wmUnloadWave(m_pregame);
	wmUnloadWave(m_levelup);
	wmUnloadWave(m_endgame);
	wmUnloadWave(m_tick);
	wmUnloadWave(m_tock);
	wmUnloadWave(m_bomb);
}

void MainGameScene::tick() {
	float ticksec = getsec() - m_statesec;

	if(m_deadlinesec != 0.0f && (m_state == GS_ZOOMINGUP || m_state == GS_GAMING)) {
		float limsec = getsec() - m_deadlinesec + LIMIT_BUFFER;

		if(m_prelimsec <= -1.5f && -1.5f < limsec) {
			// 3
			wmPlayWave(m_tick);
		}
		if(m_prelimsec <= -1.0f && -1.0f < limsec) {
			// 2
			wmPlayWave(m_tock);
		}
		if(m_prelimsec <= -0.5f && -0.5f < limsec) {
			// 1
			wmPlayWave(m_tick);
		}
		if(m_prelimsec <= 0.0f && 0.0f < limsec) {
			// bomb
			wmPlayWave(m_bomb);
		}

		m_prelimsec = limsec;
	}

	switch(m_state) {
	case GS_PREGAME:
		if(m_preticksec <= 0.0f && 0.0f < ticksec) {
			wmPlayWave(m_pregame);
		}
		if(m_preticksec <= 1.0f && 1.0f < ticksec) {
			// animated stageno++ with counter-stop.
			m_stageno = std::min(m_stageno + 1, 999);

			// choose next scene randomly.
			std::vector<SceneBuilder*> bs = g_sceneManager.builders();
			m_builder = bs[randp(bs.size())];
		}
		if(1.5f < ticksec) {
			m_gamescene = m_builder->newScene(m_level);
			if(m_gamescene->timelimit() != 0.0f) {
				m_deadlinesec = getsec() + m_gamescene->timelimit();
				m_prelimsec = -m_gamescene->timelimit();
			}
			newstate(GS_ZOOMINGUP);
			return;
		}
		break;
	case GS_ZOOMINGUP:
		m_gamescene->tick();

		if(ZOOM_DURATION < ticksec) {
			m_prevresult = SR_PLAYING;
			newstate(GS_GAMING);
			return;
		}
		break;
	case GS_GAMING:
		m_gamescene->tick();

		// scene time over?
		if(m_prevresult == SR_PLAYING && m_deadlinesec != 0.0f && m_deadlinesec < getsec()) {
			m_gamingFinSec = ticksec; // force next state now!
			m_prevresult = SR_FAILED; // timeover!
		}
		// scene finished?
		if(m_prevresult == SR_PLAYING && m_gamescene->result() != SR_PLAYING) {
			m_gamingFinSec = ticksec + 1.5f;
			if(m_deadlinesec != 0.0f) {
				m_gamingFinSec = std::min(m_gamingFinSec, ticksec + m_deadlinesec - getsec());
			}
			m_prevresult = m_gamescene->result();
		}
		// scene finished and some time forwarded?
		if(m_prevresult != SR_PLAYING && m_gamingFinSec <= ticksec) {
			newstate(GS_ZOOMINGDOWN);
			return;
		}
		break;
	case GS_ZOOMINGDOWN:
		//m_gamescene->tick();

		if(ZOOM_DURATION < ticksec) {
			delete m_gamescene;
			m_gamescene = NULL;

			m_deadlinesec = 0.0f;

			if(m_prevresult == SR_SUCCESS) {
				wmPlayWave(m_ok);
			} else if(m_prevresult == SR_FAILED) {
				wmPlayWave(m_bad);
				m_failed++; // note: incr here, or at after GS_POSTGAME?
			}

			newstate(GS_POSTGAME);
			return;
		}
		break;
	case GS_POSTGAME:
		if(2.0f < ticksec) {
			if(4 <= m_failed || (m_limitover == false && 15 <= m_stageno)) {
				// game over (whichever game clear or failed)
				wmPlayWave(m_endgame);
				newstate(GS_ENDGAME);
				return;
			} else if(m_stageno < 15 && m_stageno % 5 == 0) {
				// level up, then next stage
				wmPlayWave(m_levelup);
				newstate(GS_LEVELUP);
				return;
			} else {
				// next stage
				newstate(GS_PREGAME);
				return;
			}
		}
		break;
	case GS_LEVELUP:
		if(2.0f < ticksec) {
			m_level++;
			newstate(GS_PREGAME);
			return;
		}
		break;
	case GS_ENDGAME:
		if(5.0f < ticksec) {
			if(m_failed < 4) {
				// good end
				setNextScene(new EndrollScene(0));
			} else {
				// bad end
				setNextScene(new TitleScene(0));
			}
		}
		break;
	default:
		// BUG!
		*(char*)0=0;
		break;
	}

	m_preticksec = ticksec;
}

static void scenetitle(Painter &p, float ratio, LPCWSTR str) {
	float mag = 1.0f + std::max(0.0f, ratio) * 6.0f;
	mag = pow(mag, 4) / pow(6, 4) + 1.0f;
	Point pt = Rect(0, 0, 640, 480).center() - (g_font.bounding(mag, str) / 2).asPoint();
	g_font.draw(p, pt, mag, GLYPHRAND, str);
}

static void drawbomb(Painter &p, float deadline) {
	deadline += LIMIT_BUFFER;
	if(-3.5f < deadline) {
		Glyph *g = g_font.glyph(MGG_COUNTDOWN1 + std::max(0, static_cast<int>(ceilf(-deadline / 0.5f))));
		if(g) {
			g->draw(p, Point(0, 0), 1.0f, GLYPHRAND);
		}
	}
}

void MainGameScene::drawelevator(Painter &p, int frame) {
	Glyph *g;
	wchar_t buf[32];

	// itself
	g = g_font.glyph(MGG_ELEVATOR0 + frame);
	g->draw(p, Point(0, 0), 1.0f, GLYPHRAND);

	// floor numbers
	int frame2 = std::max(0, frame - 3);
	swprintf(buf, L"%03d", m_stageno);
	g_font.draw(p, Point(280, 30 - (frame2 * 70)), 1.0f + frame2 / 10.0f, GLYPHRAND, buf);

	// patches
	for(int i = 0; i < m_failed; i++) {
		g = g_font.glyph(MGG_PATCH1 + g_patchglyph[i]);
		Point pt = g_patchpoints[i];
		// TODO: add point by frame2
		pt = pt + ((pt - Point(320,300)) / 8 * frame2);
		g->draw(p, pt, 1.0f, GLYPHRAND);
	}

	if(m_limitover) {
		LPCWSTR str = L"☆";
		Size sz = g_font.bounding(1.0f, str);
		g_font.draw(p, Point(640 - 20, 0 + 20) - sz.onlycx().asPoint(), 1.0f, GLYPHRAND, str);
	}
}

void MainGameScene::draw(Painter &p) {
	float ticksec = getsec() - m_statesec;
	Glyph *g;

	switch(m_state) {
	case GS_PREGAME:
		drawelevator(p, 0);

		if(1.3f < ticksec) {
			/*
			float i;
			if((int)(ticksec * 2) % 2 == 0) {
				i = (ticksec * 2 - (int)(ticksec * 2));
			} else {
				i = 1 - (ticksec * 2 - (int)(ticksec * 2));
			}
			*/

			scenetitle(p, 1.0f - (ticksec - 1.3f) * 2, m_builder->sceneName());
		}

		break;
	case GS_ZOOMINGUP:
		{
			int frame = std::min(4, static_cast<int>(ticksec * ((1 / ZOOM_DURATION) * 5)));
			drawelevator(p, 1 + frame);

			// game related
			p.setViewport(g_elevatorRects[frame]);

			// game scene
			m_gamescene->draw(p);

			// bomb
			if(m_deadlinesec != 0.0f) {
				drawbomb(p, getsec() - m_deadlinesec);
			}

			p.resetViewport();

			// game scene title
			scenetitle(p, 1.0f - (0.2f + ticksec) * 2, m_builder->sceneName());
		}
		break;
	case GS_GAMING:
		{
			m_gamescene->draw(p);

			// bomb
			if(m_deadlinesec != 0.0f) {
				drawbomb(p, getsec() - m_deadlinesec);
			}

			// game scene title
			if(ticksec < 0.5f) {
				scenetitle(p, 1.0f - (0.2f + 0.5f + ticksec) * 2, m_builder->sceneName());
			}
		}
		break;
	case GS_ZOOMINGDOWN:
		{
			int frame = std::max(0, 4 - static_cast<int>(ticksec * ((1 / ZOOM_DURATION) * 5)));
			drawelevator(p, 1 + frame);

			p.setViewport(g_elevatorRects[frame]);
			m_gamescene->draw(p);
			// bomb
			if(m_deadlinesec != 0.0f) {
				drawbomb(p, getsec() - m_deadlinesec);
			}
		}
		break;
	case GS_POSTGAME:
		drawelevator(p, 0);

		switch(m_prevresult) {
		case SR_SUCCESS:
			g = g_font.glyph(MGG_OK1 + (static_cast<int>(getsec() * 2.0f) % 2));
			g->draw(p, Point(0, 0), 1.0f, GLYPHRAND);
			break;
		case SR_FAILED:
			g = g_font.glyph(MGG_BAD1 + (static_cast<int>(getsec() * 2.0f) % 2));
			g->draw(p, Point(0, 0), 1.0f, GLYPHRAND);
			break;
		default:
			// BUG!
			*(char*)0=0;
		}

		break;
	case GS_LEVELUP:
		drawelevator(p, 0);

		g = g_font.glyph(MGG_LEVELUP);
		g->draw(p, Point(0, 0), 1.0f, GLYPHRAND);

		g = g_font.glyph(MGG_LUPSHOUT1 + (static_cast<int>(getsec() * 4.0f) % 2));
		g->draw(p, Point(0, 0), 1.0f, GLYPHRAND);

		break;
	case GS_ENDGAME:
		{
			int fade = static_cast<int>(std::max(std::min(1.0f - (ticksec - 4.0f) * 2, 1.0f), 0.0f) * 255);
			PenBrushRAII pbraii(p, 3, RGB(fade, fade, fade), PS_SOLID, false, 0);

			drawelevator(p, 0);
		}
		break;
	}
}

// pump events to game scene...

void MainGameScene::button(enum MouseButton btn, Point curpos, bool isdown) {
	if(m_state == GS_ZOOMINGUP || m_state == GS_GAMING) {
		m_gamescene->button(btn, curpos, isdown);
	}
}

void MainGameScene::drag(enum MouseButton btn, Point curpos) {
	if(m_state == GS_ZOOMINGUP || m_state == GS_GAMING) {
		m_gamescene->drag(btn, curpos);
	}
}

void MainGameScene::hover(Point curpos) {
	if(m_state == GS_ZOOMINGUP || m_state == GS_GAMING) {
		m_gamescene->hover(curpos);
	}
}

void MainGameScene::key(UINT vk, bool isdown) {
	if(m_state == GS_ZOOMINGUP || m_state == GS_GAMING) {
		m_gamescene->key(vk, isdown);
	}
}

void MainGameScene::moji(TCHAR ch) {
	if(m_state == GS_ZOOMINGUP || m_state == GS_GAMING) {
		m_gamescene->moji(ch);
	}
}

void MainGameScene::newstate(enum GameState state) {
	m_state = state;
	m_statesec = getsec();
	m_preticksec = 0.0f;
}
