#include <tchar.h>
#include <windows.h>

#include <algorithm>

#include "support.h"

#include "Painter.h"

const int RIP_W = 640;
const int RIP_H = 480;

Painter::Painter(HDC hDC, bool deleteOnDtor) {
	m_hDC = hDC;

	// TODO: rip-relative pen width?
	m_hOrgPen = SelectObject(m_hDC, CreatePen(PS_SOLID, 3, RGB(255, 255, 255)));
	m_hBrush = CreateSolidBrush(RGB(255, 255, 255));
	m_hOrgBrush = SelectObject(m_hDC, m_hBrush);

	m_deleteOnDtor = deleteOnDtor;
}

Painter::~Painter() {
	DeleteObject(SelectObject(m_hDC, m_hOrgPen));
	DeleteObject(SelectObject(m_hDC, m_hOrgBrush));

	if(m_deleteOnDtor) {
		DeleteDC(m_hDC);
	}
}

void Painter::setSize(Size sz) {
	m_sz = sz;
}

void Painter::setViewport(Rect rc) {
	Point tl = rip2pt(rc.topLeft());
	Point br = rip2pt(rc.bottomRight());
	HRGN hrgn = CreateRectRgnIndirect(Rect(tl, br).prect());
	SelectClipRgn(m_hDC, hrgn);
	DeleteObject(hrgn);
}

void Painter::resetViewport() {
	SelectClipRgn(m_hDC, NULL);
}

void Painter::appendViewmask(Rect rc) {
	int ret;
	Point tl = rip2pt(rc.topLeft());
	Point br = rip2pt(rc.bottomRight());

	HRGN rgn1 = createEmptyRgn();
	if(rgn1 == NULL) *(char*)0 = 0;
	ret = GetClipRgn(m_hDC, rgn1);
	if(ret == -1) *(char*)0 = 0;

	if(ExcludeClipRect(m_hDC, tl.x(), tl.y(), br.x(), br.y()) == ERROR) *(char*)0 = 0;

	if(ret == 1) {
		// combining regions

		// get ExcludeClipRect result region
		HRGN rgn2 = createEmptyRgn();
		if(rgn2 == NULL) *(char*)0 = 0;
		if(GetClipRgn(m_hDC, rgn2) == -1) *(char*)0 = 0;

		// combine region at entry and ExcludeClipRect region
		HRGN rgn3 = createEmptyRgn();
		if(rgn3 == NULL) *(char*)0 = 0;
		if(CombineRgn(rgn3, rgn1, rgn2, RGN_AND) == ERROR) *(char*)0 = 0;

		// set that region
		if(SelectClipRgn(m_hDC, rgn3) == ERROR) *(char*)0 = 0;

		DeleteObject(rgn2);
		DeleteObject(rgn3);
	}
	DeleteObject(rgn1);
}

void Painter::line(Point p1, Point p2) {
	p1 = rip2pt(p1);
	p2 = rip2pt(p2);
	MoveToEx(m_hDC, p1.x(), p1.y(), NULL);
	LineTo(m_hDC, p2.x(), p2.y());
}

void Painter::text(Point p, LPCTSTR text) {
	// TODO: stub for bootstrapping. use Font::draw instead.
	Point pt = rip2pt(p);
	TextOut(m_hDC, pt.x(), pt.y(), text, _tcslen(text));
}

void Painter::frameRect(Rect rc) {
	Rect arect(rip2pt(rc.topLeft()), rip2pt(rc.bottomRight()));
	FrameRect(m_hDC, arect.prect(), m_hBrush);
}

void Painter::fillRect(Rect rc) {
	Rect arect(rip2pt(rc.topLeft()), rip2pt(rc.bottomRight()));
	FillRect(m_hDC, arect.prect(), m_hBrush);
}

Point Painter::rip2pt(Point p) {
	return Point(p.x() * m_sz.cx() / RIP_W, p.y() * m_sz.cy() / RIP_H);
}


PenBrushRAII::PenBrushRAII(Painter &p) : m_painter(p) {
	m_on = false;
}

PenBrushRAII::PenBrushRAII(Painter &p, int width, COLORREF color, int penstyle, bool isHatchBrush, int brushstyle) : m_painter(p) {
	m_on = false;
	on(width, color, penstyle, isHatchBrush, brushstyle);
}

PenBrushRAII::~PenBrushRAII() {
	off();
}

void PenBrushRAII::on(int width, COLORREF color, int penstyle, bool isHatchBrush, int brushstyle) {
	if(m_on) {
		return;
	}

	if(isHatchBrush) {
		m_painter.m_hBrush = CreateHatchBrush(brushstyle, color);
	} else {
		m_painter.m_hBrush = CreateSolidBrush(color);
	}
	m_hOrgSpBrush = SelectObject(m_painter.m_hDC, m_painter.m_hBrush);
	m_hOrgSpPen = SelectObject(m_painter.m_hDC, CreatePen(penstyle, width, color));

	m_on = true;
}

void PenBrushRAII::off() {
	if(!m_on) {
		return;
	}

	DeleteObject(SelectObject(m_painter.m_hDC, m_hOrgSpPen));
	DeleteObject(SelectObject(m_painter.m_hDC, m_hOrgSpBrush));
	m_painter.m_hBrush = reinterpret_cast<HBRUSH>(m_hOrgSpBrush);

	m_on = false;
}

PreserveViewrgnRAII::PreserveViewrgnRAII(Painter &p) : m_painter(p) {
	// TODO: crash code "*(char*)0 = 0;" does not crash process!!! WHY??? bcuz in ctor?

	m_orgrgn = createEmptyRgn();
	if(m_orgrgn == NULL) *(char*)0 = 0; // BUG or Resource starvation
	int ret;
	ret = GetClipRgn(m_painter.m_hDC, m_orgrgn);
	if(ret == -1) *(char*)0 = 0; // BUG or ...? 
	if(ret == 0) {
		// no region... remove all clip when dtor called.
		DeleteObject(m_orgrgn);
		m_orgrgn = NULL;
	}
}

PreserveViewrgnRAII::~PreserveViewrgnRAII() {
	// note: m_orgrgn will be NULL but it is OK.
	if(SelectClipRgn(m_painter.m_hDC, m_orgrgn) == ERROR) *(char*)0 = 0; // BUG or ...?
	if(m_orgrgn != NULL) {
		DeleteObject(m_orgrgn);
	}
}

WMPainterRAII::WMPainterRAII(HWND hWnd) {
	m_hWnd = hWnd;
	HDC hDC = BeginPaint(hWnd, &m_ps);
	m_painter = new Painter(hDC, false);
}

WMPainterRAII::~WMPainterRAII() {
	delete m_painter; // note: must deleted before calling EndPaint.
	EndPaint(m_hWnd, &m_ps);
}
