#include <tchar.h>
#include <windows.h>

#include <vector>
#include <map>

#include "support.h"
#include "zipfs/zipfs.h"
#include "wavemix/wavemix.h"
#include "zipwave.h"

#include "Scene.h"
#include "Painter.h"
#include "FontGlyph.h"

#include "Globals.h"

#include "SwordGameScene.h"


enum SwordGameGlyphs {
	SGG_BG = 0xE080,
	SGG_ENEMYUPPER = 0xE081,
	SGG_ENEMYLOWER = 0xE082,
	SGG_MYUPPER = 0xE083,
	SGG_MYLOWER = 0xE084,
	SGG_ENEMYIAI1 = 0xE085,
	SGG_ENEMYIAI2,
	SGG_ENEMYIAI3,
};

static const float BIG_THUNDER_FLASHDUR = 0.1f;
static const float SLAY_FLASHDUR = 0.3f;
static const float ENEMYA = 0.03f;
static const float MYA = 0.03f;

static SwordGameSceneBuilder g_swordGameSceneBuilder;

SwordGameScene::SwordGameScene(int level) : Scene(level) {
	m_enemyupper = Point(0, 0);
	m_myupper = Point(0, 0);
	m_enemyv = 0;
	m_myv = 0;
	m_enemyiaisec = randrf(1.0f, 7.0f);
	m_enemydelaysec = std::max(0.1f, 1.0f - (level * 0.25f));
	m_enemyslayed = false;
	m_myslayed = false;
	m_ticked = false;
	m_premature = false;
	m_slaysec = 0.0f;

	m_nextthundersec = randrf(0.0f, 10.0f);
	m_bigthundersec = 0.0f;

	m_wbgm = unzLoadWave("se_maoudamashii_element_wind01_03.wav");
	m_wbigthunder = unzLoadWave("se_maoudamashii_element_thunder05.wav");
	m_wshortthunder = unzLoadWave("se_maoudamashii_element_thunder01.wav");
	m_wiai = unzLoadWave("se_maoudamashii_element_ice05.wav");
	m_wslay = unzLoadWave("se_maoudamashii_battle05.wav");
}

SwordGameScene::~SwordGameScene() {
	wmKillWave(m_wbgm);

	wmUnloadWave(m_wbgm);
	wmUnloadWave(m_wbigthunder);
	wmUnloadWave(m_wshortthunder);
	wmUnloadWave(m_wiai);
	wmUnloadWave(m_wslay);
}

void SwordGameScene::tick() {
	float sec = getsec();

	if(!m_ticked) {
		m_ticked = true;
		wmPlayWave(m_wbgm);
	}

	if(m_presec < m_enemyiaisec && m_enemyiaisec <= sec) {
		// iai!
		wmPlayWave(m_wiai);
	}

	if(m_enemyslayed == false && m_presec < m_enemyiaisec + m_enemydelaysec && m_enemyiaisec + m_enemydelaysec <= sec) {
		// player killed
		m_myslayed = true;
		m_slaysec = sec;
		wmPlayWave(m_wslay);
		setResult(SR_FAILED);
	}

	if(m_presec < m_nextthundersec && m_nextthundersec <= sec) {
		// thunder.
		int isbig = randp(2);
		if(isbig != 0) {
			// big thunder: flash screen
			wmPlayWave(m_wbigthunder);
			m_bigthundersec = sec;
		} else {
			// short thunder: sound only
			wmPlayWave(m_wshortthunder);
		}
	}

	if(m_enemyslayed) {
		m_enemyupper += Point(-m_enemyv, m_enemyv);
		m_enemyv += ENEMYA;
	}
	if(m_myslayed) {
		m_myupper += Point(-m_myv, m_myv);
		m_myv += MYA;
	}

	m_presec = sec;
}

void SwordGameScene::draw(Painter &p) {
	float sec = getsec();
	Glyph *g;

	// BG
	g = g_font.glyph(SGG_BG);
	g->draw(p, Point(0, 0), 1.0f, GLYPHRAND);

	// enemy
	g = g_font.glyph(SGG_ENEMYUPPER);
	g->draw(p, m_enemyupper, 1.0f, GLYPHRAND);
	g = g_font.glyph(SGG_ENEMYLOWER);
	g->draw(p, Point(0, 0), 1.0f, GLYPHRAND);

	// player
	g = g_font.glyph(SGG_MYUPPER);
	g->draw(p, m_myupper, 1.0f, GLYPHRAND);
	g = g_font.glyph(SGG_MYLOWER);
	g->draw(p, Point(0, 0), 1.0f, GLYPHRAND);

	// enemy iai
	if(m_enemyiaisec <= sec && sec < m_enemyiaisec + m_enemydelaysec) {
		int frame = static_cast<int>(((sec - m_enemyiaisec) / m_enemydelaysec) * 5);
		if(3 <= frame) {
			frame = 1 - (frame - 3);
		}

		g = g_font.glyph(SGG_ENEMYIAI1 + frame);
		g->draw(p, m_enemyupper, 1.0f, GLYPHRAND);
	}

	// thunder
	if(m_bigthundersec != 0.0f && sec - m_bigthundersec < BIG_THUNDER_FLASHDUR * 3) {
		int n = static_cast<int>((sec - m_bigthundersec) / BIG_THUNDER_FLASHDUR);
		if(n % 2 == 0) {
			// flash screen
			p.fillRect(Rect(0, 0, 640, 480));
		}
	}

	// slay flash
	if(m_slaysec != 0.0f && sec - m_slaysec < SLAY_FLASHDUR) {
		// flash screen
		p.fillRect(Rect(0, 0, 640, 480));
	}
}

void SwordGameScene::button(enum MouseButton btn, Point curpos, bool isdown) {
	if(!isdown) {
		return;
	}

	if(m_premature || m_enemyslayed || m_myslayed) {
		return;
	}

	float sec = getsec();

	if(m_enemyiaisec <= sec && sec < m_enemyiaisec + m_enemydelaysec) {
		// slay success
		m_enemyslayed = true;
		m_slaysec = sec;
		wmPlayWave(m_wslay);
		setResult(SR_SUCCESS);
	} else if(sec < m_enemyiaisec) {
		// too early
		m_premature = true;
		m_enemyiaisec = sec; // iai now
		m_enemydelaysec = std::min(m_enemydelaysec, 0.3f); // slay fast
	}
}

//void SwordGameScene::drag(enum MouseButton btn, Point curpos) {}
//void SwordGameScene::hover(Point curpos) {}
