#ifdef __cplusplus
class Point;
class Rect;

class Size {
public:
	Size() {}
	Size(SIZE sz) { m_sz = sz; }
	Size(LONG cx, LONG cy) {
		m_sz.cx = cx;
		m_sz.cy = cy;
	}

	LONG cx() const { return m_sz.cx; }
	LONG cy() const { return m_sz.cy; }
	LPSIZE psize() { return &m_sz; }

	Size onlycx() { return Size(m_sz.cx, 0); }
	Size onlycy() { return Size(0, m_sz.cy); }

	Size operator +(Size sz) const { return Size(m_sz.cx + sz.cx(), m_sz.cy + sz.cy()); }
	Size operator -(Size sz) const { return Size(m_sz.cx - sz.cx(), m_sz.cy - sz.cy()); }
	Size operator *(float n) const { return Size(m_sz.cx * n, m_sz.cy * n); }
	Size operator /(float n) const { return Size(m_sz.cx / n, m_sz.cy / n); }

	Point asPoint() const; // no define here because forward declaration... define later.
	Rect rect() const; // no define here because forward declaration... define later.
private:
	SIZE m_sz;
};

class Point {
public:
	Point() {}
	Point(POINT pt) { m_pt = pt; }
	Point(LONG x, LONG y) {
		m_pt.x = x;
		m_pt.y = y;
	}

	LONG x() const { return m_pt.x; }
	LONG y() const { return m_pt.y; }
	void setx(LONG n) { m_pt.x = n; }
	void sety(LONG n) { m_pt.y = n; }
	LPPOINT ppoint() { return &m_pt; }

	Point onlyx() { return Point(m_pt.x, 0); }
	Point onlyy() { return Point(0, m_pt.y); }

	Point operator -() const { return Point(-m_pt.x, -m_pt.y); } // XXX: handling Point as a tuple is a bad idea...
	Point operator +(Point pt) const { return Point(m_pt.x + pt.x(), m_pt.y + pt.y()); }
	Point operator +(Size sz) const { return Point(m_pt.x + sz.cx(), m_pt.y + sz.cy()); }
	Point operator -(Point pt) const { return Point(m_pt.x - pt.x(), m_pt.y - pt.y()); }
	Point operator *(float n) const { return Point(m_pt.x * n, m_pt.y * n); }
	Point operator /(float n) const { return Point(m_pt.x / n, m_pt.y / n); }
	Point &operator +=(Point pt) { m_pt.x += pt.x(); m_pt.y += pt.y(); return *this; }
	Point &operator -=(Point pt) { m_pt.x -= pt.x(); m_pt.y -= pt.y(); return *this; }
	bool operator ==(Point pt) const { return m_pt.x == pt.x() && m_pt.y == pt.y(); }
	bool operator !=(Point pt) const { return !(*this == pt); }

	Rect box(int ext) const; // no define here because forward declaration... define later.
	Rect bottomright(Point br) const; // no define here because forward declaration... define later.
	Rect rect(Size sz) const; // no define here because forward declaration... define later.
	Size asSize() const { return Size(m_pt.x, m_pt.y); }
private:
	POINT m_pt;
};

class Rect {
public:
	Rect() {}
	Rect(RECT rc) { m_rc = rc; }
	Rect(Point tl, Point br) {
		m_rc.left = std::min(tl.x(), br.x());
		m_rc.top = std::min(tl.y(), br.y());
		m_rc.right = std::max(tl.x(), br.x());
		m_rc.bottom = std::max(tl.y(), br.y());
	}
	Rect(LONG l, LONG t, LONG r, LONG b) {
		m_rc.left = l;
		m_rc.top = t;
		m_rc.right = r;
		m_rc.bottom = b;
	}

	LONG left() const { return m_rc.left; }
	LONG top() const { return m_rc.top; }
	LONG right() const { return m_rc.right; }
	LONG bottom() const { return m_rc.bottom; }
	Point topLeft() const { return Point(m_rc.left, m_rc.top); }
	Point bottomRight() const { return Point(m_rc.right, m_rc.bottom); }
	void setLeft(LONG n) { m_rc.left = n; }
	void setTop(LONG n) { m_rc.top = n; }
	void setRight(LONG n) { m_rc.right = n; }
	void setBottom(LONG n) { m_rc.bottom = n; }
	LPRECT prect() { return &m_rc; }

	LONG width() const { return m_rc.right - m_rc.left; }
	LONG height() const { return m_rc.bottom - m_rc.top; }

	Point center() const { return Point(width() / 2, height() / 2); }
	Size size() const { return Size(width(), height()); }
	bool has(Point pt) const {
		return m_rc.left <= pt.x() &&
		       m_rc.top <= pt.y() &&
		       pt.x() <= m_rc.right &&
		       pt.y() <= m_rc.bottom;
	}
	Rect shrink(int n) const { return Rect(m_rc.left + n, m_rc.top + n, m_rc.right - n, m_rc.bottom - n); }
	Rect xaxis(int l, int r) const { return Rect(l, m_rc.top, r, m_rc.bottom); }
	Rect yaxis(int t, int b) const { return Rect(m_rc.left, t, m_rc.right, b); }
private:
	RECT m_rc;
};

// define pre-declared inlines here.
inline Point Size::asPoint() const { return Point(m_sz.cx, m_sz.cy); }
inline Rect Size::rect() const { return Rect(0, 0, m_sz.cx, m_sz.cy); }
inline Rect Point::box(int ext) const { return Rect(m_pt.x - ext, m_pt.y - ext, m_pt.x + ext, m_pt.y + ext); }
inline Rect Point::bottomright(Point br) const { return Rect(*this, *this + br); }
inline Rect Point::rect(Size sz) const { return Rect(*this, *this + sz); }
#endif

#ifndef _countof
#define _countof(a) (sizeof(a)/sizeof((a)[0]))
#endif

// support functions
#ifdef __cplusplus
extern "C" {
#endif
void die(LPCTSTR msg);
void debug(LPCTSTR fmt, ...);
// 0...n
int randp(int n);
// -n...+n
int randnp(int n);
// b...e
int randr(int b, int e);
// b...e in float
float randrf(float b, float e);

HRGN createEmptyRgn();
int sgn(int n);
#ifdef __cplusplus
}
#endif
